from builtins import str
from pycuda.tools import context_dependent_memoize
from neon.backends import cuda_templates

from neon.backends.cuda_templates import (_common_fp16_to_fp32,
                                          _common_round,  # for fp32_to_fp16 converter
                                          _common_max_abs,
                                          _common_kepler,
                                          _ew_types)
from neon.backends.util.source_module import SourceModule

"""
CUDA kernels for pooling layers, with support for max pooling and average pooling.
For each pooling type, there is an fprop function, a bprop function, and a bprop
for overlapping kernels. Each of the six kernels uses templating to perform dtype
conversion so it works for all data types (currently fp32 and fp16 are supported).
Additionally, there are templates for statistics collection, currently supporting
a global max abs over the output tensor which is passed in as an additional kernel
argument.
"""


def map_string2func(funcname, clss, compute_capability):
    """
    Helper function that converts string function names to function calls
    """
    if "_get_" + funcname not in globals():
        raise AttributeError("kernel type '" + funcname + "' not understood")
    return globals()["_get_" + funcname](clss, compute_capability)


# this template is used to hide variables that are only defined conditionally.
atomic_max = r"""
atomicMax(maxabs, intermediate_max);
"""

_common_divmod = r"""
__device__ __forceinline__ int div16(int numerator, int magic, int shift)
{
    int res;
    asm("vmad.s32.u32.u32 %0, %1.h0, %2.h0, 0;" : "=r"(res) : "r"(numerator), "r"(magic));
    return res >> shift;
}
__device__ __forceinline__ int mod16(int numerator, int div, int maxdiv)
{
    int res;
    asm("vmad.s32.u32.u32 %0, -%1.h0, %2.h0, %3;" : "=r"(res) : "r"(div), "r"(maxdiv), "r"(numerator));
    return res;
}
__device__ __forceinline__ int mad16(int a, int b, int c)
{
    int res;
    asm("vmad.s32.u32.u32 %0, %1.h0, %2.h0, %3;" : "=r"(res) : "r"(a), "r"(b), "r"(c));
    return res;
}
__device__ __forceinline__ int msub16(int a, int b, int c)
{
    int res;
    asm("vmad.s32.u32.u32 %0, %1.h0, %2.h0, -%3;" : "=r"(res) : "r"(a), "r"(b), "r"(c));
    return res;
}
"""

def prepare_template_vals(dtype, compute_capability, rounding=False):
    """
    Set up template code snippets that are reused across multiple kernels.
    Most are data type conversion and statistics collection related.
    """
    template_vals = dict()
    for key in ("inits", "finish", "stats_args", "mul_by_scale", "atomic_max", "cvt_out"):
        template_vals[key] = ""

    template_vals["common"] = _common_divmod

    if rounding:
        template_vals["common"] += _common_urand_gen
        template_vals["common"] += _common_round["nearest"].get(dtype, "")
        template_vals["inits"] += _init_rand_func + _init_rand_round_func
        template_vals["finish"] += _finish_rand_func
        mode = "random"
    else:
        mode = "nearest"

    template_vals["common"] += _common_round[mode].get(dtype, "")
    template_vals["common"] += _common_max_abs

    if (compute_capability[0] == 3 and compute_capability[1] < 5) or compute_capability[0] < 3:
        template_vals["common"] += _common_kepler

    template_vals["type"] = _ew_types[dtype]["type"]
    template_vals["cvt"] = _ew_types[dtype]["cvt"]

    if dtype == "f2":
        template_vals["common"] += _common_fp16_to_fp32
        template_vals["cvt_out"] = "fp32_to_fp16"
    elif dtype == "x2":
        template_vals["stats_args"] += ", int* maxabs, float scale0"
        template_vals["cvt"] = "(float)"
        template_vals["cvt_out"] = "fp32_to_int16"
        template_vals["mul_by_scale"] += "1/scale0 *"
        template_vals["atomic_max"] += atomic_max
    elif dtype == "f4":
        pass
    else:
        raise ValueError("Did not understand clss dtype " + str(dtype))

    return template_vals


# This section of the code contains templated CUDA-C code for the kernels.
@context_dependent_memoize
def _get_fprop_max(clss, compute_capability):

    code = r"""
#define FLT_MAX 3.402823466E+38F

%(common)s

__global__ void spool_fprop_max(
    const %(type)s* I, %(type)s* O, unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN, int P, int Q,
    int magic_P, int shift_P, int QN, int PQN, int MPQN,
    int pad_c, int pad_d, int pad_h, int pad_w,
    int str_c, int str_d, int str_h, int str_w,
    int S, int RS, int RST, int JRST,
    int magic_S, int shift_S,
    int magic_RS, int shift_RS, int magic_RST, int shift_RST,
    int supP, int supQ, int shlP, int maskP, int shrP,
    int shlQ, int maskQ, int shrQ, int maskN, int shrN
    %(stats_args)s
    )
{
    extern __shared__ int lut[];
    int tid = threadIdx.x;

    int q  = blockIdx.x;
    int mp = blockIdx.y;
    int k  = blockIdx.z;

    int m = mp * magic_P; m >>= shift_P;
    int p = mp - m*supP;

    // zigzag q back and forth to improve L2 cache perf
    if (p & 1)
        q = supQ - q - 1;

    // Superblock P and Q
    p = (p << shlP) + ((tid & maskP) >> shrP);
    q = (q << shlQ) + ((tid & maskQ) >> shrQ);
    int n = tid & maskN;

    int sb = tid >> shrN;

    int offset = k*MPQN + m*PQN + p*QN + mad16(q, N, n);
    I += n;
    O += offset;
    A += offset;

    float O_val = beta != 0.0f && p < P && q < Q && n < N ? %(cvt)s(__ldg(O)) : 0.0f;

    if (tid < 32)
    {
        int kj = k * str_c - pad_c;
        int mt = m * str_d - pad_d;
        int pr = p * str_h - pad_h;
        int qs = q * str_w - pad_w;

        int inc = min(maskN + 1, 32);

        int jrst = n;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int x = qs + s;
            int y = pr + r;
            int z = mt + t;
            int c = kj + j;

            bool bounds_x  = x >= 0 && x < W;
            bool bounds_y  = y >= 0 && y < H;
            bool bounds_z  = z >= 0 && z < D;
            bool bounds_c  = c >= 0 && c < C;
            bool in_bounds = bounds_x && bounds_y && bounds_z && bounds_c;

            int sliceI  = c*DHWN + z*HWN + y*WN + x*N;

            int lut_offset = mad16(sb, JRST, jrst);

            lut[lut_offset] = in_bounds ? sliceI : -1;
            jrst += inc;
        }
    }
    __syncthreads();

    int intermediate_max = 0;

    if (p < P && q < Q && n < N)
    {
        int jrst = 0;
        int argmax = 0;
        float max = -FLT_MAX;
        while (jrst < JRST)
        {
            int lut_offset = mad16(sb, JRST, jrst);

            int slice0 = lut[lut_offset + 0];
            int slice1 = lut[lut_offset + 1];
            int slice2 = lut[lut_offset + 2];
            int slice3 = lut[lut_offset + 3];

            // val needs to stay in fp32 or can't be se to FLT_MAX
            float val0 = jrst + 0 < JRST && slice0 >= 0 ? %(cvt)s(__ldg(I + slice0)) : -FLT_MAX;
            float val1 = jrst + 1 < JRST && slice1 >= 0 ? %(cvt)s(__ldg(I + slice1)) : -FLT_MAX;
            float val2 = jrst + 2 < JRST && slice2 >= 0 ? %(cvt)s(__ldg(I + slice2)) : -FLT_MAX;
            float val3 = jrst + 3 < JRST && slice3 >= 0 ? %(cvt)s(__ldg(I + slice3)) : -FLT_MAX;

            if (val0 > max) {
                max = val0;
                argmax = jrst + 0;
            }
            if (val1 > max) {
                max = val1;
                argmax = jrst + 1;
            }
            if (val2 > max) {
                max = val2;
                argmax = jrst + 2;
            }
            if (val3 > max) {
                max = val3;
                argmax = jrst + 3;
            }

            jrst += 4;
        }
        // convert back to fp to write out
        %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (max*alpha + O_val*beta));
        if (!(flags & 1)) {
            *O = temp_out;
            *A = (unsigned char)argmax;
        }

        intermediate_max = max_abs(0, temp_out);  // compute abs
    }
    intermediate_max += 0;
    %(atomic_max)s
}
"""

    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_fprop_max")
    sig = "3P 2f 44I" + ("Pf" if (clss[0] == "x") else "")
    kernel.prepare(sig)
    return kernel


@context_dependent_memoize
def _get_fprop_avg(clss, compute_capability):

    code = r"""
%(common)s

__global__ void spool_fprop_avg(
    const %(type)s* I, %(type)s* O, unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN, int P, int Q,
    int magic_P, int shift_P, int QN, int PQN, int MPQN,
    int pad_c, int pad_d, int pad_h, int pad_w,
    int str_c, int str_d, int str_h, int str_w,
    int S, int RS, int RST, int JRST,
    int magic_S, int shift_S,
    int magic_RS, int shift_RS, int magic_RST, int shift_RST,
    int supP, int supQ, int shlP, int maskP, int shrP,
    int shlQ, int maskQ, int shrQ, int maskN, int shrN
    %(stats_args)s
    )
{
    __shared__ float rcpWindowSize[32];
    extern __shared__ int lut[];

    int tid = threadIdx.x;

    int q  = blockIdx.x;
    int mp = blockIdx.y;
    int k  = blockIdx.z;

    int m = mp * magic_P; m >>= shift_P;
    int p = mp - m*supP;

    // zigzag q back and forth to improve L2 cache perf
    if (p & 1)
        q = supQ - q - 1;

    // Superblock P and Q
    p = (p << shlP) + ((tid & maskP) >> shrP);
    q = (q << shlQ) + ((tid & maskQ) >> shrQ);
    int n = tid & maskN;

    int sb = tid >> shrN;

    I += n;
    O += k*MPQN + m*PQN + p*QN + mad16(q, N, n);

    float O_val = beta != 0.0f && p < P && q < Q && n < N ? %(cvt)s(__ldg(O)) : 0.0f;

    if (tid < 32)
    {
        int kj = k * str_c - pad_c;
        int mt = m * str_d - pad_d;
        int pr = p * str_h - pad_h;
        int qs = q * str_w - pad_w;

        int inc    = min(maskN + 1, 32);
        int sbBits = 1 << min(shrN, 5);
        int sbMask = ~(-1 << sbBits) << mad16(sb, sbBits, 0);

        int window_size = 0;
        int jrst = n;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int x = qs + s;
            int y = pr + r;
            int z = mt + t;
            int c = kj + j;

            bool bounds_x  = x >= 0 && x < W;
            bool bounds_y  = y >= 0 && y < H;
            bool bounds_z  = z >= 0 && z < D;
            bool bounds_c  = c >= 0 && c < C;
            bool in_bounds = bounds_x && bounds_y && bounds_z && bounds_c;

            // Count the total valid slices
            window_size += __popc(sbMask & __ballot(in_bounds));

            int sliceI  = c*DHWN + z*HWN + y*WN + x*N;

            int lut_offset = mad16(sb, JRST, jrst);

            lut[lut_offset] = in_bounds ? sliceI : -1;
            jrst += inc;
        }

        // TODO confirm kepler OK
        unsigned int shrN_mask = (shrN < 32) ? max(0, ((1 << shrN) - 1)) : 0xffffffff;
        if((tid & shrN_mask) == 0)
            rcpWindowSize[sb] = 1.0f / (float)window_size;
    }
    __syncthreads();

    float rcp_window_size = rcpWindowSize[sb];

    int intermediate_max = 0;

    if (p < P && q < Q && n < N)
    {
        int jrst = 0;
        float sum = 0.0f;
        while (jrst < JRST)
        {
            int lut_offset = mad16(sb, JRST, jrst);

            int slice0 = lut[lut_offset + 0];
            int slice1 = lut[lut_offset + 1];
            int slice2 = lut[lut_offset + 2];
            int slice3 = lut[lut_offset + 3];

            sum += jrst + 0 < JRST && slice0 >= 0 ? %(cvt)s(__ldg(I + slice0)) : 0.0f;
            sum += jrst + 1 < JRST && slice1 >= 0 ? %(cvt)s(__ldg(I + slice1)) : 0.0f;
            sum += jrst + 2 < JRST && slice2 >= 0 ? %(cvt)s(__ldg(I + slice2)) : 0.0f;
            sum += jrst + 3 < JRST && slice3 >= 0 ? %(cvt)s(__ldg(I + slice3)) : 0.0f;

            jrst += 4;
        }

        // convert back to fp to write out
        %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (sum*rcp_window_size*alpha + O_val*beta));
        if (!(flags & 1)) {
            *O = temp_out;
        }
        // collect max abs stats
        intermediate_max = max_abs(0, temp_out); // compute abs
    }
    intermediate_max += 0;
    %(atomic_max)s
}
"""

    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("spool_fprop_avg")
    kernel.prepare("3P 2f 44I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel


@context_dependent_memoize
def _get_fprop_lrn(clss, compute_capability):
    """
    Local Response Normalization (LRN) layer.
    Implementation based on fprop_avg kernel.

    Implements the following operation:
    for each output pixel
        x' = x / response
    where the response is
        response = [1 + alpha/N * sum_neighbors x_neighbor**2 ]**beta
    so we compute the pooling output
    """
    code = r"""
%(common)s

__global__ void spool_fprop_lrn(
    const %(type)s* I, %(type)s* O, %(type)s* A,
    float alpha, float beta, float ascale, float bpower, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN, int P, int Q,
    int magic_P, int shift_P, int QN, int PQN, int MPQN,
    int pad_c, int pad_d, int pad_h, int pad_w,
    int str_c, int str_d, int str_h, int str_w,
    int S, int RS, int RST, int JRST,
    int magic_S, int shift_S,
    int magic_RS, int shift_RS, int magic_RST, int shift_RST,
    int supP, int supQ, int shlP, int maskP, int shrP,
    int shlQ, int maskQ, int shrQ, int maskN, int shrN
    %(stats_args)s
    )
{
    __shared__ float rcpWindowSize;
    extern __shared__ int lut[];

    int tid = threadIdx.x;

    // paralellism is over QMPK dimensions (output pixels and ofm's)
    int n  = tid;
    int q  = blockIdx.x;
    int mp = blockIdx.y;
    int k  = blockIdx.z;

    int m = mp * magic_P; m >>= shift_P;
    int p = mp - m*P;

    // zigzag q back and forth to improve L2 cache perf
    if (p & 1)
        q = Q - q - 1;

    const %(type)s* IonO = I;  // input pixel at output location
    I += n;
    IonO += k*MPQN + m*PQN + p*QN + q*N + n;
    O += k*MPQN + m*PQN + p*QN + q*N + n;
    A += k*MPQN + m*PQN + p*QN + q*N + n;

    float O_val = beta != 0.0f ? %(cvt)s(__ldg(O)) : 0.0f;

    if (tid < 32)
    {
        int kj = k * str_c - pad_c;
        int mt = m * str_d - pad_d;
        int pr = p * str_h - pad_h;
        int qs = q * str_w - pad_w;

        int window_size = 0;
        int jrst = tid;
        // this loop generates the LUT (same for pooling and normalization)
        while (jrst < JRST)
        {
            int j = jrst * magic_RST; j >>= shift_RST;
            int rst = jrst - j * RST;

            int t = rst * magic_RS; t >>= shift_RS;
            int rs = rst - t * RS;

            int r = rs * magic_S; r >>= shift_S;
            int s = rs - r*S;

            int x = qs + s;
            int y = pr + r;
            int z = mt + t;
            int c = kj + j;

            bool bounds_x  = x >= 0 && x < W;
            bool bounds_y  = y >= 0 && y < H;
            bool bounds_z  = z >= 0 && z < D;
            bool bounds_c  = c >= 0 && c < C;
            bool in_bounds = bounds_x && bounds_y && bounds_z && bounds_c;

            // Count the total valid slices
            window_size += __popc(__ballot(in_bounds));

            int sliceI  = c*DHWN + z*HWN + y*WN + x*N;

            lut[jrst] = in_bounds ? sliceI : -1;
            jrst += 32;
        }

        if(tid == 0)
        {
            //rcpWindowSize = 1.0f / (float)window_size;
            rcpWindowSize = (float)RST/(float)JRST;
        }
    }
    __syncthreads();

    float out = 0.0f;
    float denom;
    float sumsquare = 0.0f;
    float input = 0.0f;
    int jrst = 0;
    while (jrst < JRST)
    {
        int slice0 = lut[jrst + 0];
        int slice1 = lut[jrst + 1];
        int slice2 = lut[jrst + 2];
        int slice3 = lut[jrst + 3];

        // TODO: May not need to load all slices if they are not used.
        input =      jrst + 0 < JRST && slice0 >= 0 ? %(cvt)s(__ldg(I + slice0)) : 0.0f;
        sumsquare += jrst + 0 < JRST && slice0 >= 0 ? input * input: 0.0f;
        input =      jrst + 1 < JRST && slice1 >= 0 ? %(cvt)s(__ldg(I + slice1)) : 0.0f;
        sumsquare += jrst + 1 < JRST && slice1 >= 0 ? input * input: 0.0f;
        input =      jrst + 2 < JRST && slice2 >= 0 ? %(cvt)s(__ldg(I + slice2)) : 0.0f;
        sumsquare += jrst + 2 < JRST && slice2 >= 0 ? input * input: 0.0f;
        input =      jrst + 3 < JRST && slice3 >= 0 ? %(cvt)s(__ldg(I + slice3)) : 0.0f;
        sumsquare += jrst + 3 < JRST && slice3 >= 0 ? input * input: 0.0f;

        jrst += 4;
    }

    denom = (1 + ascale*sumsquare*rcpWindowSize);
    out = %(cvt)s(__ldg(IonO)) / powf(denom, bpower);


    // convert back to fp to write out
    %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (out*alpha + O_val*beta));

    // predicate write with no-op flag
    if (!(flags & 1)) {
        *O = temp_out;
        *A = %(cvt_out)s( %(mul_by_scale)s denom );  // write the denomiantor to address
    }

    // collect max abs stats
    int intermediate_max = max_abs(0, temp_out); // compute abs
    %(atomic_max)s
}
"""

    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_fprop_lrn")
    kernel.prepare("3P 4f 34I 10I" + ("Pf" if (clss[0] == "x") else ""))  # add superblocking parameter
    return kernel


@context_dependent_memoize
def _get_bprop_lrn_overlap(clss, compute_capability):

    code = r"""
%(common)s

union LutEntry {
    struct {
        int slice;
        int argmax;
    } data;
    int2 data2;
};

__global__ void spool_bprop_lrn_overlap(
    const %(type)s* I, const %(type)s* O, const %(type)s* E, %(type)s* delta, const %(type)s* A,
    float alpha, float beta, float ascale, float bpower, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN,
    int magic_H, int shift_H,
    int pad_w, int pad_h, int pad_d, int pad_c,
    int str_w, int str_h, int str_d, int str_c,
    int magic_str_w, int shift_str_w,
    int magic_str_h, int shift_str_h,
    int magic_str_d, int shift_str_d,
    int magic_str_c, int shift_str_c,
    int S, int R, int T, int J, int RS, int RST, int JRST,
    int magic_S, int shift_S, int magic_RS, int shift_RS,
    int magic_RST, int shift_RST,
    int Q, int P, int M, int K, int QN, int PQN, int MPQN
    %(stats_args)s  // template for "int* maxabs, float scale0"
    )
{
    extern __shared__ int2 lut[];
    __shared__ float rcpWindowSize;

    int tid = threadIdx.x;

    int n  = tid;
    int x  = blockIdx.x;
    int yz = blockIdx.y;
    int c  = blockIdx.z;

    int z = yz * magic_H; z >>= shift_H;
    int y = yz - z*H;

    // zigzag q back and forth to improve L2 cache perf
    if (y & 1)
        x = W - x - 1;

    // O E A used inside JRST loop
    O += n;  // output
    E += n;  // error
    A += n;  // denom

    // I E A used for output
    const %(type)s* E_out = E;
    const %(type)s* A_out = A;
    delta += c*DHWN + z*HWN + y*WN + x*N + n;
    I     += c*DHWN + z*HWN + y*WN + x*N + n;
    E_out += c*DHWN + z*HWN + y*WN + x*N;
    A_out += c*DHWN + z*HWN + y*WN + x*N;

    float delta_val = (beta != 0.0f) ? %(cvt)s(__ldg(delta)) : 0.0f;

    // build the lookup table
    if (tid < 32)
    {
        int kj = c - J + pad_c + 1;
        int mt = z - T + pad_d + 1;
        int pr = y - R + pad_h + 1;
        int qs = x - S + pad_w + 1;

        int jrst = tid;
        while (jrst < JRST)
        {
            int j = jrst * magic_RST; j >>= shift_RST;
            int rst = jrst - j * RST;

            int t = rst * magic_RS; t >>= shift_RS;
            int rs = rst - t * RS;

            int r = rs * magic_S; r >>= shift_S;
            int s = rs - r*S;

            int k_prime = kj + j;
            int m_prime = mt + t;
            int p_prime = pr + r;
            int q_prime = qs + s;

            int k     = k_prime * magic_str_c; k >>= shift_str_c;
            int k_mod = k_prime - k*str_c;
            bool k_bounds = k_mod == 0 && k >= 0 && k < K;

            int m     = m_prime * magic_str_d; m >>= shift_str_d;
            int m_mod = m_prime - m*str_d;
            bool m_bounds = m_mod == 0 && m >= 0 && m < M;

            int p     = p_prime * magic_str_h; p >>= shift_str_h;
            int p_mod = p_prime - p*str_h;
            bool p_bounds = p_mod == 0 && p >= 0 && p < P;

            int q     = q_prime * magic_str_w; q >>= shift_str_w;
            int q_mod = q_prime - q*str_w;
            bool q_bounds = q_mod == 0 && q >= 0 && q < Q;

            bool in_bounds = k_bounds && m_bounds && p_bounds && q_bounds;

            int j_prime = c - k_prime + pad_c;
            int t_prime = z - m_prime + pad_d;
            int r_prime = y - p_prime + pad_h;
            int s_prime = x - q_prime + pad_w;

            int sliceI  = k*MPQN + m*PQN + p*QN + q*N;
            int argmaxI = j_prime*RST + t_prime*RS + r_prime*S + s_prime;

            LutEntry entry;
            entry.data.slice  = sliceI;
            entry.data.argmax = in_bounds ? argmaxI : -1;

            lut[jrst] = entry.data2;
            jrst += 32;
        }

        if(tid == 0)
        {
            rcpWindowSize = (float)RST/(float)JRST;
        }
    }
    __syncthreads();

    int jrst = 0;
    // float out = 0.0f;
    float array_delta = 0.0f;
    int intermediate_max = 0;

    while (jrst < JRST)
    {
        LutEntry entry0;
        LutEntry entry1;
        LutEntry entry2;
        LutEntry entry3;

        entry0.data2 = lut[jrst + 0];
        entry1.data2 = lut[jrst + 1];
        entry2.data2 = lut[jrst + 2];
        entry3.data2 = lut[jrst + 3];

        array_delta += (jrst + 0 < JRST && entry0.data.argmax >= 0) ? %(cvt)s(__ldg(O + entry0.data.slice)) * %(cvt)s(__ldg(E + entry0.data.slice)) / %(cvt)s(__ldg(A + entry0.data.slice)) : 0.0f;
        array_delta += (jrst + 1 < JRST && entry1.data.argmax >= 0) ? %(cvt)s(__ldg(O + entry1.data.slice)) * %(cvt)s(__ldg(E + entry1.data.slice)) / %(cvt)s(__ldg(A + entry1.data.slice)) : 0.0f;
        array_delta += (jrst + 2 < JRST && entry2.data.argmax >= 0) ? %(cvt)s(__ldg(O + entry2.data.slice)) * %(cvt)s(__ldg(E + entry2.data.slice)) / %(cvt)s(__ldg(A + entry2.data.slice)) : 0.0f;
        array_delta += (jrst + 3 < JRST && entry3.data.argmax >= 0) ? %(cvt)s(__ldg(O + entry3.data.slice)) * %(cvt)s(__ldg(E + entry3.data.slice)) / %(cvt)s(__ldg(A + entry3.data.slice)) : 0.0f;

        jrst += 4;
    }

    array_delta = -2 * bpower * ascale * __ldg(I) * array_delta * rcpWindowSize + (__ldg(E_out) * powf(__ldg(A_out), -bpower));

    %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (array_delta*alpha + delta_val*beta));
    if (!(flags & 1)) {
        *delta = temp_out;
    }

    // compute max-abs
    intermediate_max = max_abs(intermediate_max, temp_out);  // used for abs
    %(atomic_max)s
}
"""

    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_bprop_lrn_overlap")
    kernel.prepare("5P 4f 47I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel

@context_dependent_memoize
def _get_bprop_max(clss, compute_capability):

    code = r"""

%(common)s

__global__ void spool_bprop_max(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN, int P, int Q,
    int magic_P, int shift_P, int QN, int PQN, int MPQN,
    int pad_c, int pad_d, int pad_h, int pad_w,
    int str_c, int str_d, int str_h, int str_w,
    int S, int RS, int RST, int JRST,
    int magic_S, int shift_S,
    int magic_RS, int shift_RS, int magic_RST, int shift_RST,
    int supP, int supQ, int shlP, int maskP, int shrP,
    int shlQ, int maskQ, int shrQ, int maskN, int shrN
    %(stats_args)s
    )
{
    extern __shared__ int lut[];

    int tid = threadIdx.x;

    int q  = blockIdx.x;
    int mp = blockIdx.y;
    int k  = blockIdx.z;

    int m = mp * magic_P; m >>= shift_P;
    int p = mp - m*supP;

    // zigzag q back and forth to improve L2 cache perf
    if (p & 1)
        q = supQ - q - 1;

    // Superblock P and Q
    p = (p << shlP) + ((tid & maskP) >> shrP);
    q = (q << shlQ) + ((tid & maskQ) >> shrQ);
    int n = tid & maskN;

    int sb = tid >> shrN;

    int offset = k*MPQN + m*PQN + p*QN + mad16(q, N, n);
    O += n;
    I += offset;
    A += offset;

    float delta = 0.0f;
    int argmax  = -1;
    if (p < P && q < Q && n < N)
    {
        delta  = %(cvt)s(__ldg(I));
        argmax = __ldg(A);
    }

    if (tid < 32)
    {
        int kj = k * str_c - pad_c;
        int mt = m * str_d - pad_d;
        int pr = p * str_h - pad_h;
        int qs = q * str_w - pad_w;

        int inc = min(maskN + 1, 32);

        int jrst = n;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int x = qs + s;
            int y = pr + r;
            int z = mt + t;
            int c = kj + j;

            bool bounds_x  = x >= 0 && x < W;
            bool bounds_y  = y >= 0 && y < H;
            bool bounds_z  = z >= 0 && z < D;
            bool bounds_c  = c >= 0 && c < C;
            bool in_bounds = bounds_x && bounds_y && bounds_z && bounds_c;

            int sliceI  = c*DHWN + z*HWN + y*WN + x*N;

            int lut_offset = mad16(sb, JRST, jrst);

            lut[lut_offset] = in_bounds ? sliceI : -1;
            jrst += inc;
        }
    }
    __syncthreads();

    int intermediate_max = 0;

    if (p < P && q < Q && n < N)
    {
        delta *= alpha;
        bool load_beta = beta != 0.0f;
        int jrst = 0;

        while (jrst < JRST)
        {
            int lut_offset = mad16(sb, JRST, jrst);

            int offset0 = lut[lut_offset + 0];
            int offset1 = lut[lut_offset + 1];
            int offset2 = lut[lut_offset + 2];
            int offset3 = lut[lut_offset + 3];

            // need to figure out how to write into output. Can't be float * if we write fp16
            // load fp16 from O, so it's an fp16 pointer
            %(type)s* out0 = O + offset0;
            %(type)s* out1 = O + offset1;
            %(type)s* out2 = O + offset2;
            %(type)s* out3 = O + offset3;

            bool valid0 = jrst + 0 < JRST && offset0 >= 0;
            bool valid1 = jrst + 1 < JRST && offset1 >= 0;
            bool valid2 = jrst + 2 < JRST && offset2 >= 0;
            bool valid3 = jrst + 3 < JRST && offset3 >= 0;

            // load input dtype, convert to float32.
            float beta0 = valid0 && load_beta ? %(cvt)s(__ldg(out0)) * beta : 0.0f;
            float beta1 = valid1 && load_beta ? %(cvt)s(__ldg(out1)) * beta : 0.0f;
            float beta2 = valid2 && load_beta ? %(cvt)s(__ldg(out2)) * beta : 0.0f;
            float beta3 = valid3 && load_beta ? %(cvt)s(__ldg(out3)) * beta : 0.0f;

            // convert float32 back into input format to write out
            %(type)s temp_out0 = valid0 ? %(cvt_out)s(%(mul_by_scale)s(jrst + 0 == argmax ? delta + beta0 : beta0)) : 0.0f;
            %(type)s temp_out1 = valid1 ? %(cvt_out)s(%(mul_by_scale)s(jrst + 1 == argmax ? delta + beta1 : beta1)) : 0.0f;
            %(type)s temp_out2 = valid2 ? %(cvt_out)s(%(mul_by_scale)s(jrst + 2 == argmax ? delta + beta2 : beta2)) : 0.0f;
            %(type)s temp_out3 = valid3 ? %(cvt_out)s(%(mul_by_scale)s(jrst + 3 == argmax ? delta + beta3 : beta3)) : 0.0f;

            // predicate writes with no-op flag.
            if (!(flags & 1)) {
                if (valid0) *out0 = temp_out0;
                if (valid1) *out1 = temp_out1;
                if (valid2) *out2 = temp_out2;
                if (valid3) *out3 = temp_out3;
            }
            intermediate_max = max_abs(intermediate_max, temp_out0);
            intermediate_max = max_abs(intermediate_max, temp_out1);
            intermediate_max = max_abs(intermediate_max, temp_out2);
            intermediate_max = max_abs(intermediate_max, temp_out3);

            jrst += 4;
        }
    }
    intermediate_max += 0;
    %(atomic_max)s
}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_bprop_max")
    # f = open("spool_bprop_max.cu", "w")
    # print >>f, code
    # f.close()
    kernel.prepare("3P 2f 44I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel


@context_dependent_memoize
def _get_bprop_avg(clss, compute_capability):

    code = r"""

%(common)s

__global__ void spool_bprop_avg(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN, int P, int Q,
    int magic_P, int shift_P, int QN, int PQN, int MPQN,
    int pad_c, int pad_d, int pad_h, int pad_w,
    int str_c, int str_d, int str_h, int str_w,
    int S, int RS, int RST, int JRST,
    int magic_S, int shift_S,
    int magic_RS, int shift_RS, int magic_RST, int shift_RST,
    int supP, int supQ, int shlP, int maskP, int shrP,
    int shlQ, int maskQ, int shrQ, int maskN, int shrN
    %(stats_args)s
    )
{
    __shared__ float rcpWindowSize[32];
    extern __shared__ int lut[];

    int tid = threadIdx.x;

    int q  = blockIdx.x;
    int mp = blockIdx.y;
    int k  = blockIdx.z;

    int m = mp * magic_P; m >>= shift_P;
    int p = mp - m*supP;

    // zigzag q back and forth to improve L2 cache perf
    if (p & 1)
        q = supQ - q - 1;

    // Superblock P and Q
    p = (p << shlP) + ((tid & maskP) >> shrP);
    q = (q << shlQ) + ((tid & maskQ) >> shrQ);
    int n = tid & maskN;

    int sb = tid >> shrN;

    O += n;
    I += k*MPQN + m*PQN + p*QN + mad16(q, N, n);

    float delta = 0.0f;
    if (p < P && q < Q && n < N)
        delta  = %(cvt)s(__ldg(I));

    if (tid < 32)
    {
        int kj = k * str_c - pad_c;
        int mt = m * str_d - pad_d;
        int pr = p * str_h - pad_h;
        int qs = q * str_w - pad_w;

        int inc    = min(maskN + 1, 32);
        int sbBits = 1 << min(shrN, 5);
        int sbMask = ~(-1 << sbBits) << mad16(sb, sbBits, 0);

        int window_size = 0;
        int jrst = n;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int x = qs + s;
            int y = pr + r;
            int z = mt + t;
            int c = kj + j;

            bool bounds_x  = x >= 0 && x < W;
            bool bounds_y  = y >= 0 && y < H;
            bool bounds_z  = z >= 0 && z < D;
            bool bounds_c  = c >= 0 && c < C;
            bool in_bounds = bounds_x && bounds_y && bounds_z && bounds_c;

            // Count the total valid slices
            window_size += __popc(sbMask & __ballot(in_bounds));

            int sliceI  = c*DHWN + z*HWN + y*WN + x*N;

            int lut_offset = mad16(sb, JRST, jrst);

            lut[lut_offset] = in_bounds ? sliceI : -1;
            jrst += inc;
        }
        // TODO confirm kepler OK
        unsigned int shrN_mask = (shrN < 32) ? max(0, ((1 << shrN) - 1)) : 0xffffffff;
        if((tid & shrN_mask) == 0)
            rcpWindowSize[sb] = 1.0f / (float)window_size;
    }
    __syncthreads();

    int intermediate_max = 0;

    if (p < P && q < Q && n < N)
    {
        delta *= alpha * rcpWindowSize[sb];
        bool load_beta = beta != 0.0f;
        int jrst = 0;

        while (jrst < JRST)
        {
            int lut_offset = mad16(sb, JRST, jrst);

            int offset0 = lut[lut_offset + 0];
            int offset1 = lut[lut_offset + 1];
            int offset2 = lut[lut_offset + 2];
            int offset3 = lut[lut_offset + 3];

            %(type)s* out0 = O + offset0;
            %(type)s* out1 = O + offset1;
            %(type)s* out2 = O + offset2;
            %(type)s* out3 = O + offset3;

            bool valid0 = jrst + 0 < JRST && offset0 >= 0;
            bool valid1 = jrst + 1 < JRST && offset1 >= 0;
            bool valid2 = jrst + 2 < JRST && offset2 >= 0;
            bool valid3 = jrst + 3 < JRST && offset3 >= 0;

            float beta0 = valid0 && load_beta ? %(cvt)s(__ldg(out0)) * beta : 0.0f;
            float beta1 = valid1 && load_beta ? %(cvt)s(__ldg(out1)) * beta : 0.0f;
            float beta2 = valid2 && load_beta ? %(cvt)s(__ldg(out2)) * beta : 0.0f;
            float beta3 = valid3 && load_beta ? %(cvt)s(__ldg(out3)) * beta : 0.0f;

            %(type)s temp_out0 = valid0 ? %(cvt_out)s(%(mul_by_scale)s(delta + beta0)) : 0.0f;
            %(type)s temp_out1 = valid1 ? %(cvt_out)s(%(mul_by_scale)s(delta + beta1)) : 0.0f;
            %(type)s temp_out2 = valid2 ? %(cvt_out)s(%(mul_by_scale)s(delta + beta2)) : 0.0f;
            %(type)s temp_out3 = valid3 ? %(cvt_out)s(%(mul_by_scale)s(delta + beta3)) : 0.0f;

            // predicate writes with no-op flag.
            if (!(flags & 1)) {
                if (valid0) *out0 = temp_out0;
                if (valid1) *out1 = temp_out1;
                if (valid2) *out2 = temp_out2;
                if (valid3) *out3 = temp_out3;
            }
            intermediate_max = max_abs(intermediate_max, temp_out0);
            intermediate_max = max_abs(intermediate_max, temp_out1);
            intermediate_max = max_abs(intermediate_max, temp_out2);
            intermediate_max = max_abs(intermediate_max, temp_out3);

            jrst += 4;
        }
    }
    intermediate_max += 0;
    %(atomic_max)s
}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("spool_bprop_avg")
    kernel.prepare("3P 2f 44I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel


@context_dependent_memoize
def _get_bprop_max_overlap(clss, compute_capability):

    code = r"""
%(common)s

union LutEntry {
    struct {
        int slice;
        int argmax;
    } data;
    int2 data2;
};

__global__ void spool_bprop_max_overlap(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN,
    int magic_H, int shift_H,
    int pad_w, int pad_h, int pad_d, int pad_c,
    int str_w, int str_h, int str_d, int str_c,
    int magic_str_w, int shift_str_w,
    int magic_str_h, int shift_str_h,
    int magic_str_d, int shift_str_d,
    int magic_str_c, int shift_str_c,
    int S, int R, int T, int J, int RS, int RST, int JRST,
    int magic_S, int shift_S, int magic_RS, int shift_RS,
    int magic_RST, int shift_RST,
    int Q, int P, int M, int K, int QN, int PQN, int MPQN
    %(stats_args)s  // template for "int* maxabs, float scale0"
    )
{
    int __shared__ lutSize;
    extern __shared__ int2 lut[];

    int tid = threadIdx.x;

    int n  = tid;
    int x  = blockIdx.x;
    int yz = blockIdx.y;
    int c  = blockIdx.z;

    int z = yz * magic_H; z >>= shift_H;
    int y = yz - z*H;

    // zigzag q back and forth to improve L2 cache perf
    if (y & 1)
        x = W - x - 1;

    I += n;
    A += n;
    O += c*DHWN + z*HWN + y*WN + x*N + n;

    float O_val = (beta != 0.0f) ? %(cvt)s(__ldg(O)) : 0.0f;

    int lut_size;
    if (tid < 32)
    {
        int kj = c - J + pad_c + 1;
        int mt = z - T + pad_d + 1;
        int pr = y - R + pad_h + 1;
        int qs = x - S + pad_w + 1;

        unsigned dep_thd_mask = 0xffffffff;
        dep_thd_mask >>= 32 - tid;

        lut_size = 0;

        int jrst = tid;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int k_prime = kj + j;
            int m_prime = mt + t;
            int p_prime = pr + r;
            int q_prime = qs + s;

            int  k        = div16(k_prime, magic_str_c, shift_str_c);
            int  k_mod    = mod16(k_prime, k, str_c);
            bool k_bounds = k_mod == 0 && k >= 0 && k < K;

            int  m        = div16(m_prime, magic_str_d, shift_str_d);
            int  m_mod    = mod16(m_prime, m, str_d);
            bool m_bounds = m_mod == 0 && m >= 0 && m < M;

            int  p        = div16(p_prime, magic_str_h, shift_str_h);
            int  p_mod    = mod16(p_prime, p, str_h);
            bool p_bounds = p_mod == 0 && p >= 0 && p < P;

            int  q        = div16(q_prime, magic_str_w, shift_str_w);
            int  q_mod    = mod16(q_prime, q, str_w);
            bool q_bounds = q_mod == 0 && q >= 0 && q < Q;

            bool in_bounds = k_bounds && m_bounds && p_bounds && q_bounds;

            // Get a mask of all valid slices in the warp
            unsigned ballot = __ballot(in_bounds);

            // Count the total valid slices
            unsigned warp_slices = __popc(ballot);

            if (in_bounds)
            {
                // Count all the valid slices below this threadid
                unsigned dep_thd_cnt = __popc(dep_thd_mask & ballot);

                int j_prime = c - k_prime + pad_c;
                int t_prime = z - m_prime + pad_d;
                int r_prime = y - p_prime + pad_h;
                int s_prime = x - q_prime + pad_w;

                LutEntry entry;
                entry.data.slice  = k*MPQN + m*PQN + p*QN + mad16(q, N, 0);
                entry.data.argmax = j_prime*RST + mad16(t_prime, RS, mad16(r_prime, S, s_prime));

                lut[lut_size + dep_thd_cnt] = entry.data2;
            }
            lut_size += warp_slices;
            jrst += 32;
        }
        if(tid == 0)
            lutSize = lut_size;
    }
    __syncthreads();

    lut_size = lutSize;

    int jrst = 0;
    float out = 0.0f;
    int intermediate_max = 0;

    while (jrst < lut_size)
    {
        LutEntry entry0;
        LutEntry entry1;
        LutEntry entry2;
        LutEntry entry3;

        entry0.data2 = lut[jrst + 0];
        entry1.data2 = lut[jrst + 1];
        entry2.data2 = lut[jrst + 2];
        entry3.data2 = lut[jrst + 3];

        // argmax
        int fprop_argmax0 = jrst + 0 < lut_size ? __ldg(A + entry0.data.slice) : -2;
        int fprop_argmax1 = jrst + 1 < lut_size ? __ldg(A + entry1.data.slice) : -2;
        int fprop_argmax2 = jrst + 2 < lut_size ? __ldg(A + entry2.data.slice) : -2;
        int fprop_argmax3 = jrst + 3 < lut_size ? __ldg(A + entry3.data.slice) : -2;

        out += jrst + 0 < lut_size && fprop_argmax0 == entry0.data.argmax ? %(cvt)s(__ldg(I + entry0.data.slice)) : 0.0f;
        out += jrst + 1 < lut_size && fprop_argmax1 == entry1.data.argmax ? %(cvt)s(__ldg(I + entry1.data.slice)) : 0.0f;
        out += jrst + 2 < lut_size && fprop_argmax2 == entry2.data.argmax ? %(cvt)s(__ldg(I + entry2.data.slice)) : 0.0f;
        out += jrst + 3 < lut_size && fprop_argmax3 == entry3.data.argmax ? %(cvt)s(__ldg(I + entry3.data.slice)) : 0.0f;

        jrst += 4;
    }
    %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (out*alpha + O_val*beta));
    if (!(flags & 1)) {
        *O = temp_out;
    }
    // compute max-abs
    intermediate_max = max_abs(intermediate_max, temp_out);  // used for abs
    %(atomic_max)s
}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_bprop_max_overlap")
    kernel.prepare("3P 2f 47I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel

@context_dependent_memoize
def _get_bprop_avg_overlap(clss, compute_capability):

    code = r"""

%(common)s

union LutEntry {
    struct {
        int   slice;
        float rcp_in;
    } data;
    int2 data2;
};

__device__ __forceinline__ int imin(int val1, int val2)
{
    int ret;
    asm("min.s32 %%0, %%1, %%2;" : "=r"(ret) : "r"(val1), "r"(val2));
    return ret;
}

__global__ void spool_bprop_avg_overlap(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN,
    int magic_H, int shift_H,
    int pad_w, int pad_h, int pad_d, int pad_c,
    int str_w, int str_h, int str_d, int str_c,
    int magic_str_w, int shift_str_w,
    int magic_str_h, int shift_str_h,
    int magic_str_d, int shift_str_d,
    int magic_str_c, int shift_str_c,
    int S, int R, int T, int J, int RS, int RST, int JRST,
    int magic_S, int shift_S, int magic_RS, int shift_RS,
    int magic_RST, int shift_RST,
    int Q, int P, int M, int K, int QN, int PQN, int MPQN
    %(stats_args)s  // template for "int* maxabs, float scale0"
    )
{
    int    __shared__ lutSize;
    extern __shared__ int2 lut[];

    int tid = threadIdx.x;

    int n  = tid;
    int x  = blockIdx.x;
    int yz = blockIdx.y;
    int c  = blockIdx.z;

    int z = yz * magic_H; z >>= shift_H;
    int y = yz - z*H;

    // zigzag q back and forth to improve L2 cache perf
    if (y & 1)
        x = W - x - 1;

    I += n;
    O += c*DHWN + z*HWN + y*WN + x*N + n;

    float O_val = beta != 0.0f ? %(cvt)s(__ldg(O)) : 0.0f;
    int lut_size;

    if (tid < 32)
    {
        int kj = c - J + pad_c + 1;
        int mt = z - T + pad_d + 1;
        int pr = y - R + pad_h + 1;
        int qs = x - S + pad_w + 1;

        unsigned dep_thd_mask = 0xffffffff;
        dep_thd_mask >>= 32 - tid;

        lut_size = 0;

        int jrst = tid;
        while (jrst < JRST)
        {
            int j   = div16(jrst, magic_RST, shift_RST);
            int rst = mod16(jrst, j, RST);

            int t   = div16(rst, magic_RS, shift_RS);
            int rs  = mod16(rst, t, RS);

            int r   = div16(rs, magic_S, shift_S);
            int s   = mod16(rs, r, S);

            int k_prime = kj + j;
            int m_prime = mt + t;
            int p_prime = pr + r;
            int q_prime = qs + s;

            int  k        = div16(k_prime, magic_str_c, shift_str_c);
            int  k_mod    = mod16(k_prime, k, str_c);
            bool k_bounds = k_mod == 0 && k >= 0 && k < K;

            int  m        = div16(m_prime, magic_str_d, shift_str_d);
            int  m_mod    = mod16(m_prime, m, str_d);
            bool m_bounds = m_mod == 0 && m >= 0 && m < M;

            int  p        = div16(p_prime, magic_str_h, shift_str_h);
            int  p_mod    = mod16(p_prime, p, str_h);
            bool p_bounds = p_mod == 0 && p >= 0 && p < P;

            int  q        = div16(q_prime, magic_str_w, shift_str_w);
            int  q_mod    = mod16(q_prime, q, str_w);
            bool q_bounds = q_mod == 0 && q >= 0 && q < Q;

            bool in_bounds = k_bounds && m_bounds && p_bounds && q_bounds;

            // Get a mask of all valid slices in the warp
            unsigned ballot = __ballot(in_bounds);

            // Count the total valid slices
            unsigned warp_slices = __popc(ballot);

            if (in_bounds)
            {
                // Count all the valid slices below this threadid
                unsigned dep_thd_cnt = __popc(dep_thd_mask & ballot);

                int c_left = msub16(k, str_c, pad_c);
                int z_left = msub16(m, str_d, pad_d);
                int y_left = msub16(p, str_h, pad_h);
                int x_left = msub16(q, str_w, pad_w);

                float k_in = (float)imin( imin(J + c_left, J), imin(C - c_left, J) );
                float m_in = (float)imin( imin(T + z_left, T), imin(D - z_left, T) );
                float p_in = (float)imin( imin(R + y_left, R), imin(H - y_left, R) );
                float q_in = (float)imin( imin(S + x_left, S), imin(W - x_left, S) );

                float total_in = q_in * p_in * m_in * k_in;

                float rcp_in = total_in > 0.0f ? 1.0f / total_in : 0.0f;

                LutEntry entry;
                entry.data.slice  = k*MPQN + m*PQN + p*QN + mad16(q, N, 0);
                entry.data.rcp_in = rcp_in;

                lut[lut_size + dep_thd_cnt] = entry.data2;
            }
            lut_size += warp_slices;
            jrst += 32;
        }
        if(tid==0)
            lutSize = lut_size;
    }
    __syncthreads();

    lut_size = lutSize;

    int jrst = 0;
    float out = 0.0f;
    int intermediate_max = 0;

    while (jrst < lut_size)
    {
        LutEntry entry0;
        LutEntry entry1;
        LutEntry entry2;
        LutEntry entry3;

        entry0.data2 = lut[jrst + 0];
        entry1.data2 = lut[jrst + 1];
        entry2.data2 = lut[jrst + 2];
        entry3.data2 = lut[jrst + 3];

        out += jrst + 0 < lut_size ? %(cvt)s(__ldg(I + entry0.data.slice)) * entry0.data.rcp_in : 0.0f;
        out += jrst + 1 < lut_size ? %(cvt)s(__ldg(I + entry1.data.slice)) * entry1.data.rcp_in : 0.0f;
        out += jrst + 2 < lut_size ? %(cvt)s(__ldg(I + entry2.data.slice)) * entry2.data.rcp_in : 0.0f;
        out += jrst + 3 < lut_size ? %(cvt)s(__ldg(I + entry3.data.slice)) * entry3.data.rcp_in : 0.0f;

        jrst += 4;
    }

    %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (out*alpha + O_val*beta));
    if (!(flags & 1)) {
        *O = temp_out;
    }

    // max-abs over unrolls
    intermediate_max = max_abs(intermediate_max, temp_out);  // used for abs

    %(atomic_max)s

}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("spool_bprop_avg_overlap")
    kernel.prepare("3P 2f 47I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel


@context_dependent_memoize
def _get_bprop_max_overlap_smallN(clss, compute_capability):

    code = r"""
%(common)s

union LutEntry {
    struct {
        int slice;
        int argmax;
    } data;
    int2 data2;
};

__global__ void spool_bprop_max_overlap_smallN(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN,
    int magic_H, int shift_H,
    int pad_w, int pad_h, int pad_d, int pad_c,
    int str_w, int str_h, int str_d, int str_c,
    int magic_str_w, int shift_str_w,
    int magic_str_h, int shift_str_h,
    int magic_str_d, int shift_str_d,
    int magic_str_c, int shift_str_c,
    int S, int R, int T, int J, int RS, int RST, int JRST,
    int magic_S, int shift_S, int magic_RS, int shift_RS,
    int magic_RST, int shift_RST,
    int Q, int P, int M, int K, int QN, int PQN, int MPQN,
    int supH, int supW, int shlH, int maskH, int shrH,
    int shlW, int maskW, int shrW, int maskN, int shrN, int maxLutSize
    %(stats_args)s  // template for "int* maxabs, float scale0"
    )
{
    extern __shared__ int2 lut[];

    int tid = threadIdx.x;

    int x  = blockIdx.x;
    int yz = blockIdx.y;
    int c  = blockIdx.z;

    int z = yz * magic_H; z >>= shift_H;
    int y = yz - z*supH;

    // zigzag q back and forth to improve L2 cache perf
    if (y & 1)
        x = supW - x - 1;

    // Superblock H and W
    y = (y << shlH) + ((tid & maskH) >> shrH);
    x = (x << shlW) + ((tid & maskW) >> shrW);
    int n = tid & maskN;

    int sb = tid >> shrN;

    I += n;
    A += n;
    O += c*DHWN + z*HWN + y*WN + mad16(x, N, n);

    float O_val = beta != 0.0f && y < H && x < W && n < N ? %(cvt)s(__ldg(O)) : 0.0f;

    int kj = c - J + pad_c + 1;
    int mt = z - T + pad_d + 1;
    int pr = y - R + pad_h + 1;
    int qs = x - S + pad_w + 1;

    int sbSize = maskN + 1;
    int sbBits = mad16(sb, sbSize, 0);
    unsigned sbMask = ~(0xffffffff << sbSize) << sbBits;
    unsigned dep_thd_mask = (0xffffffff >> (32 - n)) << sbBits;

    int lut_offset = mad16(sb, maxLutSize, 0);
    int lut_size = 0;
    int jrst = n;
    int JRSTend = JRST;
    if (JRSTend & maskN)
        JRSTend += sbSize - (JRSTend & maskN);

    while (jrst < JRSTend)
    {
        int j   = div16(jrst, magic_RST, shift_RST);
        int rst = mod16(jrst, j, RST);

        int t   = div16(rst, magic_RS, shift_RS);
        int rs  = mod16(rst, t, RS);

        int r   = div16(rs, magic_S, shift_S);
        int s   = mod16(rs, r, S);

        int k_prime = kj + j;
        int m_prime = mt + t;
        int p_prime = pr + r;
        int q_prime = qs + s;

        int  k        = div16(k_prime, magic_str_c, shift_str_c);
        int  k_mod    = mod16(k_prime, k, str_c);
        bool k_bounds = k_mod == 0 && k >= 0 && k < K;

        int  m        = div16(m_prime, magic_str_d, shift_str_d);
        int  m_mod    = mod16(m_prime, m, str_d);
        bool m_bounds = m_mod == 0 && m >= 0 && m < M;

        int  p        = div16(p_prime, magic_str_h, shift_str_h);
        int  p_mod    = mod16(p_prime, p, str_h);
        bool p_bounds = p_mod == 0 && p >= 0 && p < P;

        int  q        = div16(q_prime, magic_str_w, shift_str_w);
        int  q_mod    = mod16(q_prime, q, str_w);
        bool q_bounds = q_mod == 0 && q >= 0 && q < Q;

        bool in_bounds = jrst < JRST && k_bounds && m_bounds && p_bounds && q_bounds;

        // Get a mask of all valid slices in the warp
        unsigned ballot = __ballot(in_bounds);

        // Count the total valid slices in this superblock
        int sb_slices = __popc(sbMask & ballot);

        if (in_bounds)
        {
            // Count all the valid slices below this threadid
            int dep_thd_cnt = __popc(dep_thd_mask & ballot);

            int j_prime = c - k_prime + pad_c;
            int t_prime = z - m_prime + pad_d;
            int r_prime = y - p_prime + pad_h;
            int s_prime = x - q_prime + pad_w;

            int sliceI  = k*MPQN + m*PQN + p*QN + mad16(q, N, 0);
            int argmaxI = j_prime*RST + mad16(t_prime, RS, mad16(r_prime, S, s_prime));

            LutEntry entry;
            entry.data.slice  = sliceI;
            entry.data.argmax = argmaxI;

            lut[lut_offset + lut_size + dep_thd_cnt] = entry.data2;
        }
        lut_size += sb_slices;
        jrst += sbSize;
    }

    int intermediate_max = 0;

    if (y < H && x < W && n < N)
    {
        int jrst = 0;
        float out = 0.0f;

        while (jrst < maxLutSize)
        {
            LutEntry entry0;
            LutEntry entry1;
            LutEntry entry2;
            LutEntry entry3;

            lut_offset = mad16(sb, maxLutSize, jrst);

            entry0.data2 = lut[lut_offset + 0];
            entry1.data2 = lut[lut_offset + 1];
            entry2.data2 = lut[lut_offset + 2];
            entry3.data2 = lut[lut_offset + 3];

            // argmax
            int fprop_argmax0 = jrst + 0 < lut_size ? __ldg(A + entry0.data.slice) : -2;
            int fprop_argmax1 = jrst + 1 < lut_size ? __ldg(A + entry1.data.slice) : -2;
            int fprop_argmax2 = jrst + 2 < lut_size ? __ldg(A + entry2.data.slice) : -2;
            int fprop_argmax3 = jrst + 3 < lut_size ? __ldg(A + entry3.data.slice) : -2;

            out += jrst + 0 < lut_size && fprop_argmax0 == entry0.data.argmax ? %(cvt)s(__ldg(I + entry0.data.slice)) : 0.0f;
            out += jrst + 1 < lut_size && fprop_argmax1 == entry1.data.argmax ? %(cvt)s(__ldg(I + entry1.data.slice)) : 0.0f;
            out += jrst + 2 < lut_size && fprop_argmax2 == entry2.data.argmax ? %(cvt)s(__ldg(I + entry2.data.slice)) : 0.0f;
            out += jrst + 3 < lut_size && fprop_argmax3 == entry3.data.argmax ? %(cvt)s(__ldg(I + entry3.data.slice)) : 0.0f;

            jrst += 4;
        }
        %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (out*alpha + O_val*beta));
        if (!(flags & 1))
            *O = temp_out;

        // compute max-abs
        intermediate_max = max_abs(intermediate_max, temp_out);  // used for abs
    }
    intermediate_max += 0;
    %(atomic_max)s
}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    module = SourceModule(code)
    kernel = module.get_function("spool_bprop_max_overlap_smallN")
    kernel.prepare("3P 2f 58I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel


@context_dependent_memoize
def _get_bprop_avg_overlap_smallN(clss, compute_capability):

    code = r"""

%(common)s

union LutEntry {
    struct {
        int   slice;
        float rcp_in;
    } data;
    int2 data2;
};

__device__ __forceinline__ int imin(int val1, int val2)
{
    int ret;
    asm("min.s32 %%0, %%1, %%2;" : "=r"(ret) : "r"(val1), "r"(val2));
    return ret;
}

__global__ void spool_bprop_avg_overlap_smallN(
    const %(type)s* I, %(type)s* O, const unsigned char* A,
    float alpha, float beta, int flags,
    int N, int W, int H, int D, int C,
    int WN, int HWN, int DHWN,
    int magic_H, int shift_H,
    int pad_w, int pad_h, int pad_d, int pad_c,
    int str_w, int str_h, int str_d, int str_c,
    int magic_str_w, int shift_str_w,
    int magic_str_h, int shift_str_h,
    int magic_str_d, int shift_str_d,
    int magic_str_c, int shift_str_c,
    int S, int R, int T, int J, int RS, int RST, int JRST,
    int magic_S, int shift_S, int magic_RS, int shift_RS,
    int magic_RST, int shift_RST,
    int Q, int P, int M, int K, int QN, int PQN, int MPQN,
    int supH, int supW, int shlH, int maskH, int shrH,
    int shlW, int maskW, int shrW, int maskN, int shrN, int maxLutSize
    %(stats_args)s  // template for "int* maxabs, float scale0"
    )
{
    extern __shared__ int2 lut[];

    int tid = threadIdx.x;

    int x  = blockIdx.x;
    int yz = blockIdx.y;
    int c  = blockIdx.z;

    int z = yz * magic_H; z >>= shift_H;
    int y = yz - z*supH;

    // zigzag q back and forth to improve L2 cache perf
    if (y & 1)
        x = supW - x - 1;

    // Superblock H and W
    y = (y << shlH) + ((tid & maskH) >> shrH);
    x = (x << shlW) + ((tid & maskW) >> shrW);
    int n = tid & maskN;

    int sb = tid >> shrN;

    I += n;
    O += c*DHWN + z*HWN + y*WN + mad16(x, N, n);

    float O_val = beta != 0.0f && y < H && x < W && n < N ? %(cvt)s(__ldg(O)) : 0.0f;

    int kj = c - J + pad_c + 1;
    int mt = z - T + pad_d + 1;
    int pr = y - R + pad_h + 1;
    int qs = x - S + pad_w + 1;

    int sbSize = maskN + 1;
    int sbBits = mad16(sb, sbSize, 0);
    unsigned sbMask = ~(0xffffffff << sbSize) << sbBits;
    unsigned dep_thd_mask = (0xffffffff >> (32 - n)) << sbBits;

    int lut_offset = mad16(sb, maxLutSize, 0);
    int lut_size = 0;
    int jrst = n;
    int JRSTend = JRST;
    if (JRSTend & maskN)
        JRSTend += sbSize - (JRSTend & maskN);

    while (jrst < JRSTend)
    {
        int j   = div16(jrst, magic_RST, shift_RST);
        int rst = mod16(jrst, j, RST);

        int t   = div16(rst, magic_RS, shift_RS);
        int rs  = mod16(rst, t, RS);

        int r   = div16(rs, magic_S, shift_S);
        int s   = mod16(rs, r, S);

        int k_prime = kj + j;
        int m_prime = mt + t;
        int p_prime = pr + r;
        int q_prime = qs + s;

        int  k        = div16(k_prime, magic_str_c, shift_str_c);
        int  k_mod    = mod16(k_prime, k, str_c);
        bool k_bounds = k_mod == 0 && k >= 0 && k < K;

        int  m        = div16(m_prime, magic_str_d, shift_str_d);
        int  m_mod    = mod16(m_prime, m, str_d);
        bool m_bounds = m_mod == 0 && m >= 0 && m < M;

        int  p        = div16(p_prime, magic_str_h, shift_str_h);
        int  p_mod    = mod16(p_prime, p, str_h);
        bool p_bounds = p_mod == 0 && p >= 0 && p < P;

        int  q        = div16(q_prime, magic_str_w, shift_str_w);
        int  q_mod    = mod16(q_prime, q, str_w);
        bool q_bounds = q_mod == 0 && q >= 0 && q < Q;

        bool in_bounds = jrst < JRST && k_bounds && m_bounds && p_bounds && q_bounds;

        // Get a mask of all valid slices in the warp
        unsigned ballot = __ballot(in_bounds);

        // Count the total valid slices in this superblock
        int sb_slices = __popc(sbMask & ballot);

        if (in_bounds)
        {
            // Count all the valid slices below this threadid
            int dep_thd_cnt = __popc(dep_thd_mask & ballot);

            int c_left = msub16(k, str_c, pad_c);
            int z_left = msub16(m, str_d, pad_d);
            int y_left = msub16(p, str_h, pad_h);
            int x_left = msub16(q, str_w, pad_w);

            float k_in = (float)imin( imin(J + c_left, J), imin(C - c_left, J) );
            float m_in = (float)imin( imin(T + z_left, T), imin(D - z_left, T) );
            float p_in = (float)imin( imin(R + y_left, R), imin(H - y_left, R) );
            float q_in = (float)imin( imin(S + x_left, S), imin(W - x_left, S) );

            float total_in = q_in * p_in * m_in * k_in;

            LutEntry entry;
            entry.data.slice  = k*MPQN + m*PQN + p*QN + mad16(q, N, 0);
            entry.data.rcp_in = total_in > 0.0f ? 1.0f / total_in : 0.0f;

            lut[lut_offset + lut_size + dep_thd_cnt] = entry.data2;
        }
        lut_size += sb_slices;
        jrst += sbSize;
    }

    int intermediate_max = 0;

    if (y < H && x < W && n < N)
    {
        int jrst = 0;
        float out = 0.0f;

        while (jrst < maxLutSize)
        {
            LutEntry entry0;
            LutEntry entry1;
            LutEntry entry2;
            LutEntry entry3;

            lut_offset = mad16(sb, maxLutSize, jrst);

            entry0.data2 = lut[lut_offset + 0];
            entry1.data2 = lut[lut_offset + 1];
            entry2.data2 = lut[lut_offset + 2];
            entry3.data2 = lut[lut_offset + 3];

            out += jrst + 0 < lut_size ? %(cvt)s(__ldg(I + entry0.data.slice)) * entry0.data.rcp_in : 0.0f;
            out += jrst + 1 < lut_size ? %(cvt)s(__ldg(I + entry1.data.slice)) * entry1.data.rcp_in : 0.0f;
            out += jrst + 2 < lut_size ? %(cvt)s(__ldg(I + entry2.data.slice)) * entry2.data.rcp_in : 0.0f;
            out += jrst + 3 < lut_size ? %(cvt)s(__ldg(I + entry3.data.slice)) * entry3.data.rcp_in : 0.0f;

            jrst += 4;
        }
        %(type)s temp_out = %(cvt_out)s( %(mul_by_scale)s (out*alpha + O_val*beta));
        if (!(flags & 1))
            *O = temp_out;

        // max-abs over unrolls
        intermediate_max = max_abs(intermediate_max, temp_out);  // used for abs
    }
    intermediate_max += 0;
    %(atomic_max)s

}
"""
    template_vals = prepare_template_vals(clss, compute_capability)
    code = code % template_vals
    # f = open("pool.cu", "w")
    # print >>f, code
    # f.close()
    module = SourceModule(code, options=["--use_fast_math"])
    kernel = module.get_function("spool_bprop_avg_overlap_smallN")
    kernel.prepare("3P 2f 58I" + ("Pf" if (clss[0] == "x") else ""))
    return kernel



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------
from __future__ import division
from builtins import str
from neon import NervanaObject
import numpy as np
from collections import Counter
from neon import logger as neon_logger


class Cost(NervanaObject):

    """
    Base class for cost functions that are used during training.

    Child classes can either implement the below `__call__` and `bprop` methods, or alternatively
    define `self.func` and self.funcgrad`. The latter is typically used for code
    compactness when the operations can be fit into a lambda function.
    """

    def __call__(self, y, t):
        """
        Applies the cost function

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the cost
        """
        return self.func(y, t)

    def bprop(self, y, t):
        """
        Computes the derivative of the cost function

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the derivative of the cost function
        """
        return self.funcgrad(y, t)


class CrossEntropyBinary(Cost):

    """
    Binary cross-entropy cost.

    The binary cross-entropy cost is used when the labels have two classes: 0 and 1.
    The cost is computed as :math:`C = \sum -t\log(y)-(1-t)\log(1-y)`, where :math:`t` is
    the target label and :math:`y` is the network output.

    Note:
    The backpropagation assumes that this cost is coupled with an output layer
    that uses the Logistic() activation function. This allows for a shortcut in
    the deriviate that saves computation.
    """

    def __init__(self, scale=1):
        """
        Args:
            scale (float, optional): Amount by which to scale the backpropagated error (default: 1)
        """
        self.scale = scale

    def __call__(self, y, t):
        """
        Returns the binary cross entropy cost.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the binary cross entropy cost
        """
        assert y.shape == t.shape, "CrossEntropy requires network output shape to match targets"
        return self.be.sum(self.be.safelog(1 - y) * (t - 1) - self.be.safelog(y) * t, axis=0)

    def bprop(self, y, t):
        """
        Returns the derivative of the binary cross entropy cost.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the (mean) shortcut derivative of the binary entropy
                    cost function ``(y - t) / y.shape[1]``
        """
        return self.scale * (y - t)


class CrossEntropyMulti(Cost):

    """
    Multi-class cross-entropy cost.

    The multi-class cross-entropy cost is used when the labels have multiple classes.
    The cost is computed as :math:`C = \sum -t*\log(y)`, where :math:`t` is
    the target label and :math:`y` is the network output.

    The target labels :math:`t` are expected to be in an one-hot encoding. By default,
    the natural logarithm is used, but a cost that returns bits instead (e.g. log base 2)
    can also be specified with the ``usebits`` argument.

    Note:
        The back-propogation assumes that this cost is coupled with an output layer
        that uses the Softmax() activation function. This allows for a shortcut in
        the deriviate that saves computation.
    """

    def __init__(self, scale=1, usebits=False):
        """
        Args:
            scale (float, optional): scale factor for the backpropagated error (default: 1)
            usebits (boolean, optional): Display costs in bits (default: False)
        """
        super(CrossEntropyMulti, self).__init__()
        self.usebits = usebits
        self.scale = scale
        self.logscale = np.float(1. / np.log(2.0) if usebits else 1.)

    def __call__(self, y, t):
        """
        Returns the multiclass cross entropy cost

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the multiclass cross entropy cost
        """
        if y.shape != t.shape:
            raise ValueError((
                "CrossEntropy requires network output shape to match "
                "targets. Network output shape was {} and targets shape "
                "was {}"
            ).format(y.shape, t.shape))
        return (self.be.sum(-t * self.logscale * self.be.safelog(y), axis=0))

    def bprop(self, y, t):
        """
        Returns the derivative of the multiclass cross entropy cost.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            OpTree: Returns the (mean) shortcut derivative of the multiclass
            entropy cost function ``(y - t) / y.shape[1]``
        """
        return self.logscale * self.scale * (y - t)


class SumSquared(Cost):

    """
    Total Squared Error cost function. Computes :math:`\sum_i (y_i-t_i)^2`.


    """

    def __init__(self):
        """
        Define the cost function and its gradient as lambda functions.
        """
        self.func = lambda y, t: self.be.sum(self.be.square(y - t), axis=0) / 2.
        self.funcgrad = lambda y, t: (y - t)


class MeanSquared(Cost):

    """
    Average Squared Error cost function. Computes :math:`\\frac{1}{N}\\sum_i (y_i-t_i)^2`.
    """

    def __init__(self):
        """
        Define the cost function and its gradient as lambda functions.
        """
        self.func = lambda y, t: self.be.mean(self.be.square(y - t), axis=0) / 2.
        self.funcgrad = lambda y, t: (y - t) / y.shape[0]


class SmoothL1Loss(Cost):

    """
    Smooth L1 cost function.

    The L1 loss is less sensitive to outliers than the L2 loss.
    See `Girshick 2015 <http://arxiv.org/pdf/1504.08083v2.pdf>`__. This
    cost is used for training object localization models such as Fast-RCNN.
    """

    def smoothL1(self, x):
        """
        Returns the Smooth-L1 cost
        """
        return (0.5 * self.be.square(x) * self._sigma2 * (self.be.absolute(x) < 1/self._sigma2) +
                (self.be.absolute(x) - 0.5/self._sigma2) * (self.be.absolute(x) >= 1/self._sigma2))

    def smoothL1grad(self, x):
        """
        Returns the gradient of the Smooth-L1 cost.
        """
        return (x * self._sigma2 * (self.be.absolute(x) < 1/self._sigma2) +
                self.be.sgn(x) * (self.be.absolute(x) >= 1/self._sigma2))

    def __init__(self, sigma=1.0):
        """
        Define the cost function and its gradient as lambda functions.
        """
        self.sigma = sigma
        self._sigma2 = self.be.square(sigma)
        self.func = lambda y, t: self.be.sum(self.smoothL1(y - t), axis=0)
        self.funcgrad = lambda y, t: self.smoothL1grad(y - t)


class SquareHingeLoss(Cost):

    """
    Applies the square hinge loss cost function
    """

    def squarehinge(self, y, t):
        t = 2 * t - 1
        return self.be.mean(self.be.square(self.be.maximum(self.margin - t * y, 0)), axis=0)

    def squarehingegrad(self, y, t):
        t = 2 * t - 1
        return -2 * t * self.be.maximum(self.margin - t * y, 0)/float(y.shape[0])

    def __init__(self, margin=1):
        """
        Initialize the square hinge loss cost function
        """
        self.margin = margin
        self.func = lambda y, t: self.squarehinge(y, t)
        self.funcgrad = lambda y, t: self.squarehingegrad(y, t)


class Metric(Cost):

    """
    Base class for Metrics. Metrics are quantities not used during training
    for the back-propogration but are useful to compute and display to check
    on progress.

    For example, when training on image classification network,
    we may want to use the Cross-entropy cost to train the weights, but display
    the misclassification rate metric.
    """

    def __call__(self, y, t):
        """
        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            float: Returns the metric
        """
        raise NotImplementedError()


class MultiMetric(Metric):
    """
    A wrapper Metric which can be used with Tree models which have more than
    one output.  Tree models have tuples of tensors, one tensor per output.
    Wrapping a Metric with a MultiMetric ensures that the metric sees only one
    of those tensors in the output tuple instead of all of them.
    """

    def __init__(self, metric, index):
        """
        Args:
            metric (Metric): Metric to apply in this multi-output context
            index (integer): The index into the model's output tuple to apply
                             the metric to
        """
        self.metric = metric
        self.index = index

    def __call__(self, y, t, *args, **kwargs):
        """
        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            numpy array : Returns the log loss  metric in numpy array,
                         [LogLoss]
        """
        return self.metric(y[self.index], y[self.index], *args, **kwargs)

    def __getattr__(self, key):
        return getattr(self.metric, key)


class LogLoss(Metric):

    """
    LogLoss metric.

    Computes :math:`\\log\\left(\\sum y*t\\right)`.
    """

    def __init__(self):
        self.correctProbs = self.be.iobuf(1)
        self.metric_names = ['LogLoss']

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            numpy array : Returns the log loss  metric in numpy array,
                         [LogLoss]
        """
        self.correctProbs[:] = self.be.sum(y * t, axis=0)
        self.correctProbs[:] = -self.be.safelog(self.correctProbs)
        return np.array(self.correctProbs.get()[:, calcrange].mean())


class TopKMisclassification(Metric):

    """
    Multiple misclassification metrics.

    Computes the LogLoss metric, the Top-1 Misclassification rate, and the Top-K
    misclassification rate.
    """

    def __init__(self, k):
        """
        Arguments:
            k (integer): Number of guesses to allow.
        """
        self.correctProbs = self.be.iobuf(1)
        self.top1 = self.be.iobuf(1)
        self.topk = self.be.iobuf(1)

        self.k = k
        self.metric_names = ['LogLoss', 'Top1Misclass', 'Top' + str(k) + 'Misclass']

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Returns a numpy array of metrics for: LogLoss, Top-1, and Top-K.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y
            calcrange (slice, optional): Slice of data used for the metric (default: all)

        Returns:
            numpy array : Returns the metrics in a numpy array:
                          [LogLoss, Top 1 misclass, Top k misclass]
        """
        be = self.be
        self.correctProbs[:] = be.sum(y * t, axis=0)
        nSlots = self.k - be.sum((y > self.correctProbs), axis=0)
        nEq = be.sum(y == self.correctProbs, axis=0)
        self.topk[:] = 1. - (nSlots > 0) * ((nEq <= nSlots) * (1 - nSlots / nEq) + nSlots / nEq)
        self.top1[:] = 1. - (be.max(y, axis=0) == self.correctProbs) / nEq
        self.correctProbs[:] = -be.safelog(self.correctProbs)
        return np.array((self.correctProbs.get()[:, calcrange].mean(),
                         self.top1.get()[:, calcrange].mean(),
                         self.topk.get()[:, calcrange].mean()))


class Misclassification(Metric):

    """
    Misclassification error metric.
    """

    def __init__(self, steps=1):
        """
        Initialize the metric.
        """
        self.preds = self.be.iobuf((1, steps), persist_values=False)
        self.hyps = self.be.iobuf((1, steps), persist_values=False)
        self.outputs = self.preds  # Contains per record metric
        self.metric_names = ['Top1Misclass']

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Returns the misclassification error metric

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            float: Returns the metric
        """
        # convert back from onehot and compare
        self.preds[:] = self.be.argmax(y, axis=0)
        self.hyps[:] = self.be.argmax(t, axis=0)
        self.outputs[:] = self.be.not_equal(self.preds, self.hyps)

        return self.outputs.get()[:, calcrange].mean()


class Accuracy(Metric):

    """
    Accuracy metric (correct rate).

    """

    def __init__(self):
        self.preds = self.be.iobuf(1)
        self.hyps = self.be.iobuf(1)
        self.outputs = self.preds  # Contains per record metric
        self.metric_names = ['Accuracy']

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Returns the accuracy.

        Args:
            y (Tensor or OpTree): Output of previous layer or model
            t (Tensor or OpTree): True targets corresponding to y

        Returns:
            float: Returns the metric
        """
        # convert back from onehot and compare
        self.preds[:] = self.be.argmax(y, axis=0)
        self.hyps[:] = self.be.argmax(t, axis=0)
        self.outputs[:] = self.be.equal(self.preds, self.hyps)

        return self.outputs.get()[:, calcrange].mean()


class PrecisionRecall(Metric):
    """
    Precision and Recall metrics.

    Typically used in a conjunction with a multi-classification model.
    """
    def __init__(self, num_classes, binarize=False, epsilon=1e-6):
        """
        Arguments:
            num_classes (int): Number of different output classes.
            binarize (bool, optional): If True will attempt to convert the model
                                       outputs to a one-hot encoding (in place).
                                       Defaults to False.
            epsilon (float, optional): Smoothing to apply to avoid division by zero.
                                       Defaults to 1e-6.
        """
        self.outputs = self.be.empty((num_classes, 2))
        self.token_stats = self.be.empty((num_classes, 3))
        self.metric_names = ['Precision', 'Recall']
        if binarize:
            self.bin_buf = self.be.iobuf(1, dtype=np.int32)
        else:
            self.bin_buf = None
        self.eps = epsilon

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Returns a numpy array with the precision and recall metrics.

        Args:
            y (Tensor or OpTree): Output of previous layer or model (we assume
                                  already binarized, or you need to ensure
                                  binarize is True during construction).
            t (Tensor or OpTree): True targets corresponding to y (we assume
                                  already binarized)

        Returns:
            ndarray: The class averaged precision (item 0) and recall (item
                     1) values.  Per-class statistics remain in self.outputs.
        """
        if self.bin_buf is not None:
            self.be.argmax(y, axis=0, out=self.bin_buf)
            y[:] = self.be.onehot(self.bin_buf, axis=0)
        # True positives
        self.token_stats[:, 0] = self.be.sum(y * t, axis=1)

        # Prediction
        self.token_stats[:, 1] = self.be.sum(y, axis=1)

        # Targets
        self.token_stats[:, 2] = self.be.sum(t, axis=1)

        # Precision
        self.outputs[:, 0] = self.token_stats[:, 0] / (self.token_stats[:, 1] + self.eps)

        # Recall
        self.outputs[:, 1] = self.token_stats[:, 0] / (self.token_stats[:, 2] + self.eps)

        return self.outputs.get().mean(axis=0)


class ObjectDetection(Metric):

    """
    The object detection metric includes object label accuracy, and
    bounding box regression.
    """

    def __init__(self):
        self.metric_names = ['Accuracy', 'SmoothL1Loss']
        self.label_ind = 0
        self.bbox_ind = 1

    def smoothL1(self, x):
        """
        Returns the Smooth L1 cost.
        """
        return (0.5 * self.be.square(x) * (self.be.absolute(x) < 1) +
                (self.be.absolute(x) - 0.5) * (self.be.absolute(x) >= 1))

    def __call__(self, y, t, calcrange=slice(0, None)):
        """
        Returns a numpy array with the accuracy and the Smooth-L1 metrics.

        Args:
            y (Tensor or OpTree): Output of a model like Fast-RCNN model with 2 elements:
                                    1. class label: (# classes, # batchsize for ROIs)
                                    2. object bounding box (# classes * 4, # bacthsize for ROIs)
            t (Tensor or OpTree): True targets corresponding to y, with 2 elements:
                                    1. class labels: (# classes, # batchsize for ROIs)
                                        1.1 class labels
                                                    (# classes, # batchsize for ROIs)
                                        1.2 class labels mask
                                                    (# classes, # batchsize for ROIs)
                                    2. object bounding box and mask, where mask will indicate the
                                        real object to detect other than the background objects
                                        2.1 object bounding box
                                                    (# classes * 4, # bacthsize for ROIs)
                                        2.2 object bounding box mask
                                                    (# classes * 4, # bacthsize for ROIs)

        Returns:
            numpy ary : Returns the metrics in numpy array [Label Accuracy, Bounding Box Smooth-L1]
        """
        t_bb = t[self.bbox_ind][0]
        t_bb_mask = t[self.bbox_ind][1]
        y_bb = y[self.bbox_ind]

        self.detectionMetric = self.be.empty((1, t_bb.shape[1]))
        self.detectionMetric[:] = self.be.sum(self.smoothL1(y_bb * t_bb_mask - t_bb), axis=0)

        if isinstance(t[self.label_ind], tuple):
            t_lbl = t[self.label_ind][0] * t[self.label_ind][1]
            y_lbl = y[self.label_ind] * t[self.label_ind][1]
        else:
            t_lbl = t[self.label_ind]
            y_lbl = y[self.label_ind]

        self.preds = self.be.empty((1, y_lbl.shape[1]))
        self.hyps = self.be.empty((1, t_lbl.shape[1]))
        self.labelMetric = self.be.empty((1, y_lbl.shape[1]))

        self.preds[:] = self.be.argmax(y_lbl, axis=0)
        self.hyps[:] = self.be.argmax(t_lbl, axis=0)
        self.labelMetric[:] = self.be.equal(self.preds, self.hyps)

        return np.array((self.labelMetric.get()[:, calcrange].mean(),
                         self.detectionMetric.get()[:, calcrange].mean()))


class BLEUScore(Metric):
    """
    Compute BLEU score metric
    """

    def __init__(self, unk='<unk>'):
        self.metric_names = ['BLEU']
        self.end_token = '.'
        self.unk_symbol = unk

    def __call__(self, y, t, N=4, brevity_penalty=False, lower_case=True):
        """
        Args:
            y (list): list of predicted sentences
            t (list): list of reference sentences where each element is a list
                      of multiple references
            N (int, optional): compute all ngram modified precisions up to this N
            brevity_penalty (bool, optional): if True, use brevity penalty
            lower_case (bool, optional): if True, convert all words to lower case
        """

        y_list = list(y)
        t_list = list(t)

        if lower_case:
            for ii, sent in enumerate(y_list):
                y_list[ii] = sent.lower()

        # convert all sentences to lists of words
        for ii, sent in enumerate(y_list):
            y_list[ii] = sent.strip(self.end_token).split()

        for ii, refs in enumerate(t_list):
            tmp = []
            for ref in refs:
                tmp += [ref.split()]
            t_list[ii] = tmp

        def ngram_counts(sentence, counts, N):
            for n in range(1, N+1):
                num = len(sentence) - n + 1     # number of n-grams
                for jj in range(num):
                    ngram = ' '.join(sentence[jj:jj+n])
                    ngram = repr(n) + ' ' + ngram
                    counts[ngram] += 1

        # compute ngram counts
        totals = np.zeros(N)    # ngram counts over all candidates
        correct = np.zeros(N)   # correct ngrams (compared to max over references)
        len_translation, len_reference = (0, 0)
        for ii, sent in enumerate(y_list):
            counts_ref_max = Counter()    # maximum ngram count over all references for an example
            # count ngrams in candidate sentence
            counts_cand = Counter()
            ngram_counts(sent, counts_cand, N)
            # process reference sentences
            closest_diff, closest_len = (float("inf"), float("inf"))
            for ref in t_list[ii]:
                counts_ref = Counter()
                # find closest length of reference sentence for current example
                diff = abs(len(sent) - len(ref))
                if diff < closest_diff:
                    closest_len = len(ref)
                elif diff == closest_diff:
                    closest_len = min(closest_len, len(ref))
                # compute all ngram counts up to specified n=N for this reference
                ngram_counts(ref, counts_ref, N)
                for ngram, count in counts_ref.items():
                    if counts_ref_max[ngram] < count:
                        counts_ref_max[ngram] = count
            len_reference += closest_len
            len_translation += len(sent)
            for ngram, count in counts_cand.items():
                n = int(ngram[0])
                ind = n - 1
                totals[ind] += count
                # only match if there are no UNK
                if ngram.find(self.unk_symbol) == -1:
                    r = counts_ref_max[ngram]
                    c = count if r >= count else r
                    correct[ind] += c

        # calculate bleu scores
        precision = correct/totals + 0.0000001

        if (brevity_penalty and len_translation < len_reference):
            bp = np.exp(1-float(len_reference)/len_translation)
        else:
            bp = 1.0

        logprec = np.log(precision)
        self.bleu_n = [100*bp*np.exp(sum(logprec[:nn+1])/(nn+1)) for nn in range(N)]

        neon_logger.display("Bleu scores: " + " ".join([str(np.round(f, 2)) for f in self.bleu_n]))

        return self.bleu_n[-1]


class GANCost(Cost):

    """
    Discriminator cost for a Generative Adversarial Network
    The Discriminator cost is a packaged cross-entropy where the inputs with label 0
    and the inputs with label 1 are passed in separately. It takes the form
    :math:`C = \log (y_data) + \log (1 - y_noise)` where :math:`y_data` are the fprop
    outputs of the data minibatch, and :math:`y_noise` are the outputs of the generator-
    discriminator stack on a noise batch.
    """
    def __init__(self, scale=1., cost_type="dis", func='modified'):
        """
        Args:
            scale (float, optional): Amount by which to scale the backpropagated error (default: 1)
            cost_type (string): select discriminator cost "dis" or generator cost "gen"
            cost_func (string): cost function: choice from "original", "modified" and "wasserstein"
                                (Goodfellow et al. 2014, Arjovski et al. 2017)
        """
        self.scale = scale
        self.cost_type = cost_type
        self.func = func
        err_str = "Illegal GAN cost type, can only be: gen or dis"
        assert self.cost_type in ['dis', 'gen']
        err_str = "Unsupported GAN cost function, supported: original, modified, wasserstein"
        assert self.func in ['original', 'modified', 'wasserstein'], err_str
        self.one_buf = self.be.iobuf(1)
        self.one_buf.fill(1)

    def __call__(self, y_data, y_noise, cost_type='dis'):
        """
        Returns the discriminator cost. Note sign flip of the discriminator
        cost relative to Goodfellow et al. 2014 so we can minimize the cost
        rather than maximizing discriminiation.
        Args:
            y_data (Tensor or OpTree): Output of the data minibatch
            y_noise (Tensor or OpTree): Output of noise minibatch
            cost_type (str): 'dis' (default), 'dis_data', 'dis_noise' or 'gen'
        Returns:
            OpTree: discriminator or generator cost, controlled by cost_type
        """
        assert y_data.shape == y_noise.shape, "Noise and data output shape mismatch"
        if self.func == 'original':
            cost_dis_data = -self.be.safelog(y_data)
            cost_dis_noise = -self.be.safelog(1-y_noise)
            cost_gen = -self.be.safelog(y_noise)
        elif self.func == 'modified':
            cost_dis_data = -self.be.safelog(y_data)
            cost_dis_noise = -self.be.safelog(1-y_noise)
            cost_gen = self.be.safelog(1-y_noise)
        elif self.func == 'wasserstein':
            cost_dis_data = y_data
            cost_dis_noise = -y_noise
            cost_gen = y_noise

        if cost_type == 'dis':
            return self.be.mean(cost_dis_data + cost_dis_noise, axis=0)
        elif cost_type == 'dis_data':
            return self.be.mean(cost_dis_data, axis=0)
        elif cost_type == 'dis_noise':
            return self.be.mean(cost_dis_noise, axis=0)
        elif cost_type == 'gen':
            return self.be.mean(cost_gen, axis=0)

    def bprop_noise(self, y_noise):
        """
        Derivative of the discriminator cost wrt. y_noise.
        """
        if self.func in ['original', 'modified']:
            return self.scale / (1. - y_noise)
        elif self.func == 'wasserstein':
            return -self.scale * self.one_buf

    def bprop_data(self, y_data):
        """
        Derivative of the discriminator cost wrt. y_data.
        """
        if self.func in ['original', 'modified']:
            return self.scale * (-1. / y_data)
        elif self.func == 'wasserstein':
            return self.scale * self.one_buf

    def bprop_generator(self, y_noise):
        """
        Derivative of the generator cost wrt. y_noise.
        """
        if self.func == 'original':
            return self.scale / (y_noise - 1.)
        elif self.func == 'modified':
            return -self.scale / y_noise
        elif self.func == 'wasserstein':
            return self.scale * self.one_buf



# ----------------------------------------------------------------------------
# Copyright 2014-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

from neon.data.dataiterator import NervanaDataIterator, ArrayIterator
from neon.data.hdf5iterator import HDF5Iterator, HDF5IteratorOneHot, HDF5IteratorAutoencoder
from neon.data.datasets import Dataset
from neon.data.text import Text, Shakespeare, PTB, HutterPrize, IMDB, SICK
from neon.data.questionanswer import BABI, QA
from neon.data.ticker import Ticker, CopyTask, RepeatCopyTask, PrioritySortTask
from neon.data.imagecaption import ImageCaption, ImageCaptionTest, Flickr8k, Flickr30k, Coco
from neon.data.image import MNIST, CIFAR10, DUMMY



from __future__ import print_function
from neon import logger as neon_logger

try:
    from pycuda.compiler import SourceModule as pycSourceModule
except ImportError:
    neon_logger.display("PyCUDA module not found")
from neon.util.persist import get_cache_dir


class SourceModule(pycSourceModule):
    """
    Just provides a wrapper that sets the default cache dir to a neon-specific location
    """
    def __init__(self, source, nvcc="nvcc", options=None, keep=False, no_extern_c=False,
                 arch=None, code=None, cache_dir=None, include_dirs=[]):

        if cache_dir is None:
            cache_dir = get_cache_dir(['kernels', 'pycuda'])

        return super(SourceModule, self).__init__(source, nvcc, options, keep, no_extern_c,
                                                  arch, code, cache_dir, include_dirs)



# ----------------------------------------------------------------------------
# Copyright 2015-2016 Nervana Systems Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ----------------------------------------------------------------------------

# Imports should not be a requirement for building documentation
try:
    from bokeh.plotting import figure
    from bokeh.palettes import brewer
    from bokeh.models import Range1d
    from bokeh.embed import components
    from jinja2 import Template
except ImportError:
    pass


def x_label(epoch_axis):
    """
    Get the x axis label depending on the boolean epoch_axis.

    Arguments:
        epoch_axis (bool): If true, use Epoch, if false use Minibatch

    Returns:
        str: "Epoch" or "Minibatch"
    """
    return "Epoch" if epoch_axis else "Minibatch"


def cost_fig(cost_data, plot_height, plot_width, epoch_axis=True):
    """
    Generate a figure with lines for each element in cost_data.

    Arguments:
        cost_data (list): Cost data to plot
        plot_height (int): Plot height
        plot_width (int): Plot width
        epoch_axis (bool, optional): If true, use Epoch, if false use Minibatch

    Returns:
        bokeh.plotting.figure: cost_data figure
    """

    fig = figure(plot_height=plot_height,
                 plot_width=plot_width,
                 title="Cost",
                 x_axis_label=x_label(epoch_axis),
                 y_axis_label="Cross Entropy Error (%)")

    # Spectral palette supports 3 - 11 distinct colors
    num_colors_required = len(cost_data)
    assert num_colors_required <= 11, "Insufficient colors in predefined palette."
    colors = list(brewer["Spectral"][max(3, len(cost_data))])
    if num_colors_required < 3:
        # manually adjust pallette for better contrast
        colors[0] = brewer["Spectral"][6][0]
        if num_colors_required == 2:
            colors[1] = brewer["Spectral"][6][-1]

    for name, x, y in cost_data:
        fig.line(x, y, legend=name, color=colors.pop(0), line_width=2)
    return fig


def hist_fig(hist_data, plot_height, plot_width, x_range=None, epoch_axis=True):
    """
    Generate a figure with an image plot for hist_data, bins on the Y axis and
    time on the X axis.

    Arguments:
        hist_data (tuple): Hist data to plot
        plot_height (int): Plot height
        plot_width (int): Plot width
        x_range (tuple, optional): (start, end) range for x
        epoch_axis (boolm optional): If true, use Epoch, if false use Minibatch

    Returns:
        bokeh.plotting.figure: hist_data figure
    """
    name, hdata, dh, dw, bins, offset = hist_data
    if x_range is None:
        x_range = (0, dw)
    fig = figure(plot_height=plot_height,
                 plot_width=plot_width,
                 title=name,
                 x_axis_label=x_label(epoch_axis),
                 x_range=x_range,
                 y_range=(offset, offset + bins))
    fig.image(image=[hdata], x=[0], y=[offset], dw=[dw], dh=[dh], palette="Spectral11")
    return fig


def image_fig(data, h, w, x_range, y_range, plot_size):
    """
    Helper function to generate a figure

    Arguments:
        data (int): data to plot
        h (int): height
        w (int): width
        x_range (tuple, optional): (start, end) range for x
        y_range (tuple, optional): (start, end) range for y
        plot_size (int): plot size

    Returns:
        bokeh.plotting.figure: Generated figure
    """
    fig = figure(x_range=x_range, y_range=y_range,
                 plot_width=plot_size, plot_height=plot_size,
                 toolbar_location=None)
    fig.image_rgba([data], x=[0], y=[0], dw=[w], dh=[h])
    fig.axis.visible = None
    fig.min_border = 0
    return fig


def deconv_figs(layer_name, layer_data, fm_max=8, plot_size=120):
    """
    Helper function to generate deconv visualization figures

    Arguments:
        layer_name (str): Layer name
        layer_data (list): Layer data to plot
        fm_max (int): Max layers to process
        plot_size (int, optional): Plot size

    Returns:
        tuple if vis_keys, img_keys, fig_dict
    """
    vis_keys = dict()
    img_keys = dict()
    fig_dict = dict()

    for fm_num, (fm_name, deconv_data, img_data) in enumerate(layer_data):

        if fm_num >= fm_max:
            break

        img_h, img_w = img_data.shape
        x_range = Range1d(start=0, end=img_w)
        y_range = Range1d(start=0, end=img_h)
        img_fig = image_fig(img_data, img_h, img_w, x_range, y_range, plot_size)
        deconv_fig = image_fig(deconv_data, img_h, img_w, x_range, y_range, plot_size)

        title = "{}_fmap_{:04d}".format(layer_name, fm_num)
        vis_keys[fm_num] = "vis_" + title
        img_keys[fm_num] = "img_" + title

        fig_dict[vis_keys[fm_num]] = deconv_fig
        fig_dict[img_keys[fm_num]] = img_fig

    return vis_keys, img_keys, fig_dict


def deconv_summary_page(filename, cost_data, deconv_data):
    """
    Generate an HTML page with a Deconv visualization

    Arguments:
        filename: Output filename
        cost_data (list): Cost data to plot
        deconv_data (tuple): deconv data to plot
    """
    fig_dict = dict()

    cost_key = "cost_plot"
    fig_dict[cost_key] = cost_fig(cost_data, 300, 533, epoch_axis=True)

    vis_keys = dict()
    img_keys = dict()
    for layer, layer_data in deconv_data:
        lyr_vis_keys, lyr_img_keys, lyr_fig_dict = deconv_figs(layer, layer_data, fm_max=4)
        vis_keys[layer] = lyr_vis_keys
        img_keys[layer] = lyr_img_keys
        fig_dict.update(lyr_fig_dict)

    script, div = components(fig_dict)

    template = Template('''
<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="utf-8">
        <title>{{page_title}}</title>
        <style> div{float: left;} </style>
        <link rel="stylesheet"
              href="http://cdn.pydata.org/bokeh/release/bokeh-0.9.0.min.css"
              type="text/css" />
        <script type="text/javascript"
                src="http://cdn.pydata.org/bokeh/release/bokeh-0.9.0.min.js"></script>
        {{ script }}
    </head>
    <body>
    <div id=cost_plot style="width:100%; padding:10px">
      {{ div[cost_key]}}
    </div>

    {% for layer in sorted_layers %}
        <div id=Outer{{layer}} style="padding:20px">
        <div id={{layer}} style="background-color: #C6FFF1; padding:10px">
        Layer {{layer}}<br>
        {% for fm in vis_keys[layer].keys() %}
            <div id={{fm}} style="padding:10px">
            Feature Map {{fm}}<br>
            {{ div[vis_keys[layer][fm]] }}
            {{ div[img_keys[layer][fm]] }}
            </div>
        {% endfor %}
        </div>
        </div>

        <br><br>
    {% endfor %}
    </body>
</html>
''')

    with open(filename, 'w') as htmlfile:
        htmlfile.write(template.render(page_title="Deconv Visualization", script=script,
                                       div=div, cost_key=cost_key, vis_keys=vis_keys,
                                       img_keys=img_keys,
                                       sorted_layers=sorted(vis_keys)))
